@@ -11,8 +11,11 @@
 #if defined(__GLASGOW_HASKELL__) && __GLASGOW_HASKELL__ >= 704
 # define LANGUAGE_DeriveGeneric
 {-# LANGUAGE DeriveGeneric #-}
+#else
+{-# LANGUAGE TypeFamilies #-}
 #endif
 
+{-# OPTIONS_GHC -fno-warn-deprecations #-}
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 -- |
@@ -34,6 +37,9 @@
 --
 --   * 'Quasi' for 'ReaderT', 'WriterT', 'StateT', and 'RWST'.
 --
+--   * 'Typeable' for 'Lift', 'NameIs', 'Ppr', 'PprM', 'Q', 'Quasi',
+--   'QuasiQuoter', and 'TExp'
+--
 -- More recent versions of template-haskell, particularly 2.10 (GHC
 -- 7.10), provide these instances.  However, in order to support older
 -- versions you should import this module.
@@ -65,6 +71,7 @@ import Numeric.Natural (Natural)
 import Language.Haskell.TH.Ppr
 # if MIN_VERSION_template_haskell(2,3,0)
 import Language.Haskell.TH.PprLib
+import Language.Haskell.TH.Quote
 # endif
 # if MIN_VERSION_template_haskell(2,4,0) && !(MIN_VERSION_template_haskell(2,8,0))
 import Language.Haskell.TH.Syntax.Internals
@@ -82,11 +89,13 @@ import Data.Word (Word)
 # endif
 
 # if MIN_VERSION_template_haskell(2,3,0) && defined(LANGUAGE_DeriveDataTypeable)
-import Data.Data (Data, Typeable)
+import Data.Data hiding (Fixity(..))
 # endif
 
-# if defined(LANGUAGE_DeriveGeneric)
+# if __GLASGOW_HASKELL__ > 702
 import GHC.Generics (Generic)
+# else
+import qualified Generics.Deriving.TH as Generic (deriveAll)
 # endif
 #endif
 
@@ -162,6 +171,12 @@ deriving instance Ord Stmt
 deriving instance Ord Strict
 deriving instance Ord Type
 
+# if defined(LANGUAGE_DeriveDataTypeable)
+deriving instance Typeable  NameIs
+deriving instance Typeable1 PprM
+deriving instance Typeable1 Q
+# endif
+
 # if defined(LANGUAGE_DeriveGeneric)
 deriving instance Generic Body
 deriving instance Generic Callconv
@@ -185,6 +200,29 @@ deriving instance Generic Safety
 deriving instance Generic Stmt
 deriving instance Generic Strict
 deriving instance Generic Type
+# else
+$(Generic.deriveAll ''Body)
+$(Generic.deriveAll ''Callconv)
+$(Generic.deriveAll ''Clause)
+$(Generic.deriveAll ''Con)
+$(Generic.deriveAll ''Dec)
+$(Generic.deriveAll ''Exp)
+$(Generic.deriveAll ''Fixity)
+$(Generic.deriveAll ''FixityDirection)
+$(Generic.deriveAll ''Foreign)
+$(Generic.deriveAll ''FunDep)
+$(Generic.deriveAll ''Guard)
+$(Generic.deriveAll ''Info)
+$(Generic.deriveAll ''Lit)
+$(Generic.deriveAll ''Match)
+$(Generic.deriveAll ''Name)
+$(Generic.deriveAll ''NameSpace)
+$(Generic.deriveAll ''Pat)
+$(Generic.deriveAll ''Range)
+$(Generic.deriveAll ''Safety)
+$(Generic.deriveAll ''Stmt)
+$(Generic.deriveAll ''Strict)
+$(Generic.deriveAll ''Type)
 # endif
 
 # if MIN_VERSION_template_haskell(2,3,0)
@@ -205,10 +243,14 @@ deriving instance Show Loc
 #  if defined(LANGUAGE_DeriveDataTypeable)
 deriving instance Data Loc
 deriving instance Typeable Loc
+
+deriving instance Typeable QuasiQuoter
 #  endif
 
 #  if defined(LANGUAGE_DeriveGeneric)
 deriving instance Generic Loc
+#  else
+$(Generic.deriveAll ''Loc)
 #  endif
 # endif
 
@@ -226,6 +268,14 @@ deriving instance Generic PkgName
 deriving instance Generic Pragma
 deriving instance Generic Pred
 deriving instance Generic TyVarBndr
+#  else
+$(Generic.deriveAll ''FamFlavour)
+$(Generic.deriveAll ''ModName)
+$(Generic.deriveAll ''OccName)
+$(Generic.deriveAll ''PkgName)
+$(Generic.deriveAll ''Pragma)
+$(Generic.deriveAll ''Pred)
+$(Generic.deriveAll ''TyVarBndr)
 #  endif
 
 #  if !(MIN_VERSION_template_haskell(2,8,0))
@@ -235,6 +285,9 @@ deriving instance Ord Kind
 #   if defined(LANGUAGE_DeriveGeneric)
 deriving instance Generic InlineSpec
 deriving instance Generic Kind
+#   else
+$(Generic.deriveAll ''InlineSpec)
+$(Generic.deriveAll ''Kind)
 #   endif
 #  endif
 # endif
@@ -245,6 +298,8 @@ deriving instance Ord ClassInstance
 
 #  if defined(LANGUAGE_DeriveGeneric)
 deriving instance Generic ClassInstance
+#  else
+$(Generic.deriveAll ''ClassInstance)
 #  endif
 # endif
 
@@ -267,6 +322,12 @@ deriving instance Generic Phases
 deriving instance Generic RuleBndr
 deriving instance Generic RuleMatch
 deriving instance Generic TyLit
+#  else
+$(Generic.deriveAll ''Inline)
+$(Generic.deriveAll ''Phases)
+$(Generic.deriveAll ''RuleBndr)
+$(Generic.deriveAll ''RuleMatch)
+$(Generic.deriveAll ''TyLit)
 #  endif
 # endif
 
@@ -279,6 +340,10 @@ deriving instance Ord ModuleInfo
 deriving instance Ord Role
 deriving instance Ord TySynEqn
 
+#  if defined(LANGUAGE_DeriveDataTypeable)
+deriving instance Typeable TExp
+#  endif
+
 #  if defined(LANGUAGE_DeriveGeneric)
 deriving instance Generic AnnLookup
 deriving instance Generic AnnTarget
@@ -286,6 +351,13 @@ deriving instance Generic Module
 deriving instance Generic ModuleInfo
 deriving instance Generic Role
 deriving instance Generic TySynEqn
+#  else
+$(Generic.deriveAll ''AnnLookup)
+$(Generic.deriveAll ''AnnTarget)
+$(Generic.deriveAll ''Module)
+$(Generic.deriveAll ''ModuleInfo)
+$(Generic.deriveAll ''Role)
+$(Generic.deriveAll ''TySynEqn)
 #  endif
 # else
 deriving instance Show ModName
@@ -302,91 +374,105 @@ instance Applicative PprM where
 #endif
 
 instance Quasi m => Quasi (ReaderT r m) where
-  qNewName          = MTL.lift . qNewName
-  qReport a b       = MTL.lift $ qReport a b
-  qRecover m1 m2    = ReaderT $ \ r -> runReaderT m1 r `qRecover` runReaderT m2 r
-  qReify            = MTL.lift . qReify
-  qLocation         = MTL.lift qLocation
-  qRunIO            = MTL.lift . qRunIO
+  qNewName            = MTL.lift . qNewName
+  qReport a b         = MTL.lift $ qReport a b
+  qRecover m1 m2      = ReaderT $ \ r -> runReaderT m1 r `qRecover` runReaderT m2 r
+  qReify              = MTL.lift . qReify
+  qLocation           = MTL.lift qLocation
+  qRunIO              = MTL.lift . qRunIO
 #if MIN_VERSION_template_haskell(2,7,0)
   qReifyInstances a b = MTL.lift $ qReifyInstances a b
-  qLookupName a b   = MTL.lift $ qLookupName a b
-  qAddDependentFile = MTL.lift . qAddDependentFile
-#if MIN_VERSION_template_haskell(2,9,0)
-  qReifyRoles       = MTL.lift . qReifyRoles
-  qReifyAnnotations = MTL.lift . qReifyAnnotations
-  qReifyModule      = MTL.lift . qReifyModule
-  qAddTopDecls      = MTL.lift . qAddTopDecls
-  qAddModFinalizer  = MTL.lift . qAddModFinalizer
-  qGetQ             = MTL.lift qGetQ
-  qPutQ             = MTL.lift . qPutQ
-#endif
+  qLookupName a b     = MTL.lift $ qLookupName a b
+  qAddDependentFile   = MTL.lift . qAddDependentFile
+# if MIN_VERSION_template_haskell(2,9,0)
+  qReifyRoles         = MTL.lift . qReifyRoles
+  qReifyAnnotations   = MTL.lift . qReifyAnnotations
+  qReifyModule        = MTL.lift . qReifyModule
+  qAddTopDecls        = MTL.lift . qAddTopDecls
+  qAddModFinalizer    = MTL.lift . qAddModFinalizer
+  qGetQ               = MTL.lift qGetQ
+  qPutQ               = MTL.lift . qPutQ
+# endif
+#elif MIN_VERSION_template_haskell(2,5,0)
+  qClassInstances a b = MTL.lift $ qClassInstances a b
 #endif
 
 instance (Quasi m, Monoid w) => Quasi (WriterT w m) where
-  qNewName          = MTL.lift . qNewName
-  qReport a b       = MTL.lift $ qReport a b
-  qRecover m1 m2    = WriterT $ runWriterT m1 `qRecover` runWriterT m2
-  qReify            = MTL.lift . qReify
-  qLocation         = MTL.lift qLocation
-  qRunIO            = MTL.lift . qRunIO
+  qNewName            = MTL.lift . qNewName
+  qReport a b         = MTL.lift $ qReport a b
+  qRecover m1 m2      = WriterT $ runWriterT m1 `qRecover` runWriterT m2
+  qReify              = MTL.lift . qReify
+  qLocation           = MTL.lift qLocation
+  qRunIO              = MTL.lift . qRunIO
 #if MIN_VERSION_template_haskell(2,7,0)
   qReifyInstances a b = MTL.lift $ qReifyInstances a b
-  qLookupName a b   = MTL.lift $ qLookupName a b
-  qAddDependentFile = MTL.lift . qAddDependentFile
-#if MIN_VERSION_template_haskell(2,9,0)
-  qReifyRoles       = MTL.lift . qReifyRoles
-  qReifyAnnotations = MTL.lift . qReifyAnnotations
-  qReifyModule      = MTL.lift . qReifyModule
-  qAddTopDecls      = MTL.lift . qAddTopDecls
-  qAddModFinalizer  = MTL.lift . qAddModFinalizer
-  qGetQ             = MTL.lift qGetQ
-  qPutQ             = MTL.lift . qPutQ
-#endif
+  qLookupName a b     = MTL.lift $ qLookupName a b
+  qAddDependentFile   = MTL.lift . qAddDependentFile
+# if MIN_VERSION_template_haskell(2,9,0)
+  qReifyRoles         = MTL.lift . qReifyRoles
+  qReifyAnnotations   = MTL.lift . qReifyAnnotations
+  qReifyModule        = MTL.lift . qReifyModule
+  qAddTopDecls        = MTL.lift . qAddTopDecls
+  qAddModFinalizer    = MTL.lift . qAddModFinalizer
+  qGetQ               = MTL.lift qGetQ
+  qPutQ               = MTL.lift . qPutQ
+# endif
+#elif MIN_VERSION_template_haskell(2,5,0)
+  qClassInstances a b = MTL.lift $ qClassInstances a b
 #endif
 
 instance Quasi m => Quasi (StateT s m) where
-  qNewName          = MTL.lift . qNewName
-  qReport a b       = MTL.lift $ qReport a b
-  qRecover m1 m2    = StateT $ \ s -> runStateT m1 s `qRecover` runStateT m2 s
-  qReify            = MTL.lift . qReify
-  qLocation         = MTL.lift qLocation
-  qRunIO            = MTL.lift . qRunIO
+  qNewName            = MTL.lift . qNewName
+  qReport a b         = MTL.lift $ qReport a b
+  qRecover m1 m2      = StateT $ \ s -> runStateT m1 s `qRecover` runStateT m2 s
+  qReify              = MTL.lift . qReify
+  qLocation           = MTL.lift qLocation
+  qRunIO              = MTL.lift . qRunIO
 #if MIN_VERSION_template_haskell(2,7,0)
   qReifyInstances a b = MTL.lift $ qReifyInstances a b
-  qLookupName a b   = MTL.lift $ qLookupName a b
-  qAddDependentFile = MTL.lift . qAddDependentFile
-#if MIN_VERSION_template_haskell(2,9,0)
-  qReifyRoles       = MTL.lift . qReifyRoles
-  qReifyAnnotations = MTL.lift . qReifyAnnotations
-  qReifyModule      = MTL.lift . qReifyModule
-  qAddTopDecls      = MTL.lift . qAddTopDecls
-  qAddModFinalizer  = MTL.lift . qAddModFinalizer
-  qGetQ             = MTL.lift qGetQ
-  qPutQ             = MTL.lift . qPutQ
-#endif
+  qLookupName a b     = MTL.lift $ qLookupName a b
+  qAddDependentFile   = MTL.lift . qAddDependentFile
+# if MIN_VERSION_template_haskell(2,9,0)
+  qReifyRoles         = MTL.lift . qReifyRoles
+  qReifyAnnotations   = MTL.lift . qReifyAnnotations
+  qReifyModule        = MTL.lift . qReifyModule
+  qAddTopDecls        = MTL.lift . qAddTopDecls
+  qAddModFinalizer    = MTL.lift . qAddModFinalizer
+  qGetQ               = MTL.lift qGetQ
+  qPutQ               = MTL.lift . qPutQ
+# endif
+#elif MIN_VERSION_template_haskell(2,5,0)
+  qClassInstances a b = MTL.lift $ qClassInstances a b
 #endif
 
 instance (Quasi m, Monoid w) => Quasi (RWST r w s m) where
-  qNewName          = MTL.lift . qNewName
-  qReport a b       = MTL.lift $ qReport a b
-  qRecover m1 m2    = RWST $ \ r s -> runRWST m1 r s `qRecover` runRWST m2 r s
-  qReify            = MTL.lift . qReify
-  qLocation         = MTL.lift qLocation
-  qRunIO            = MTL.lift . qRunIO
+  qNewName            = MTL.lift . qNewName
+  qReport a b         = MTL.lift $ qReport a b
+  qRecover m1 m2      = RWST $ \ r s -> runRWST m1 r s `qRecover` runRWST m2 r s
+  qReify              = MTL.lift . qReify
+  qLocation           = MTL.lift qLocation
+  qRunIO              = MTL.lift . qRunIO
 #if MIN_VERSION_template_haskell(2,7,0)
   qReifyInstances a b = MTL.lift $ qReifyInstances a b
-  qLookupName a b   = MTL.lift $ qLookupName a b
-  qAddDependentFile = MTL.lift . qAddDependentFile
-#if MIN_VERSION_template_haskell(2,9,0)
-  qReifyRoles       = MTL.lift . qReifyRoles
-  qReifyAnnotations = MTL.lift . qReifyAnnotations
-  qReifyModule      = MTL.lift . qReifyModule
-  qAddTopDecls      = MTL.lift . qAddTopDecls
-  qAddModFinalizer  = MTL.lift . qAddModFinalizer
-  qGetQ             = MTL.lift qGetQ
-  qPutQ             = MTL.lift . qPutQ
+  qLookupName a b     = MTL.lift $ qLookupName a b
+  qAddDependentFile   = MTL.lift . qAddDependentFile
+# if MIN_VERSION_template_haskell(2,9,0)
+  qReifyRoles         = MTL.lift . qReifyRoles
+  qReifyAnnotations   = MTL.lift . qReifyAnnotations
+  qReifyModule        = MTL.lift . qReifyModule
+  qAddTopDecls        = MTL.lift . qAddTopDecls
+  qAddModFinalizer    = MTL.lift . qAddModFinalizer
+  qGetQ               = MTL.lift qGetQ
+  qPutQ               = MTL.lift . qPutQ
+# endif
+#elif MIN_VERSION_template_haskell(2,5,0)
+  qClassInstances a b = MTL.lift $ qClassInstances a b
 #endif
+
+#if MIN_VERSION_base(4,7,0) && defined(LANGUAGE_DeriveDataTypeable) && __GLASGOW_HASKELL__ < 710
+deriving instance Typeable Lift
+deriving instance Typeable Ppr
+deriving instance Typeable Quasi
 #endif
 
 $(reifyManyWithoutInstances ''Lift [''Info, ''Loc] (const True) >>=
